const DEBUG = 5;
var showNamespaces = false;
var xmlNamespaces = {};
var strings;
var oldSelectedDataSource;

// File constants copied from file-utils.js
const MODE_RDONLY   = 0x01;
const MODE_WRONLY   = 0x02;
const MODE_RDWR     = 0x04;
const MODE_CREATE   = 0x08;
const MODE_APPEND   = 0x10;
const MODE_TRUNCATE = 0x20;
const MODE_SYNC     = 0x40;
const MODE_EXCL     = 0x80;

var RDF = Components
    .classes['@mozilla.org/rdf/rdf-service;1']
    .getService()
    .QueryInterface(Components.interfaces.nsIRDFService);

var RDFC = Components
    .classes["@mozilla.org/rdf/container-utils;1"]
    .getService(Components.interfaces.nsIRDFContainerUtils);

var promptService = Components
    .classes["@mozilla.org/embedcomp/prompt-service;1"]
    .getService(Components.interfaces.nsIPromptService);

function RDFInspector(aDataSource) {
    this.open(aDataSource);
}
RDFInspector.prototype = {
    mDS            : null,  // Stores the datasource itself
    mURI           : null,  // Shortcut to the datasource URI; read-only
    mDSType        : null,  // Shortcut to the datasource type; read-only
    mLabel         : null,  // User-settable label for this datasource; read/write
    mChildren      : null,  // Child datasources, if applicable; read/write
    mOpen          : false, // Flag indicating if this datasource is open; read-only
    mPaused        : false, // Flag indicating we are watching this datasource; read/write
    mLoading       : false, // Flag indicating this datasource is still loading; read-only
    debug          : 0,
    mRDFXML        : null,  // Stores the RDF/XML serialization of this datasource
    mRDFXMLTainted : false, // Indicates if the RDF/XML stored in the above field is old

    // Placeholders for internal objects
    mRDFObserver : null,
    mXMLSink     : null,

    // Storage structures for statements
    mStatements : null,

    get label() { return this.mLabel },
    set label(aLabel) { this.mLabel = aLabel },

    get URI() { return this.mURI },
    get type() { return this.mDSType },
    get length() { return this.mStatements.length },

    get isOpen() { return this.mOpen },

    get isPaused() { return this.mPaused },
    set isPaused(aPaused) {
        this.mPaused = aPaused;
    },

    get hasChildren() { return this.mChildren.length > 0 },
    get children()    { return this.mChildren },
    get isContainer() { return this.type == "nsIRDFCompositeDataSource" },

    serialize : function() {
        if (this.mRDFXMLTainted == false) {
            return this.mRDFXML;
        }

        var ds = this.cloneDataSource();
        var outputStream = {
            data: "",
            close : function() {},
            flush : function() {},
            write : function(buffer, count){
                this.data += buffer;
                return count;
            },
            writeFrom : function(stream, count) {},
            isNonBlocking: false
        };
        var serializer = Components
            .classes["@mozilla.org/rdf/xml-serializer;1"]
            .createInstance(Components.interfaces.nsIRDFXMLSerializer);
        serializer.init(ds);
        var atomService = Components
            .classes["@mozilla.org/atom-service;1"]
            .createInstance(Components.interfaces.nsIAtomService);
        for (var idx = 0; idx < xmlNamespaces.length; idx++) {
            var prefixAtom = atomService.getAtom(xmlNamespaces[idx].prefix);
            serializer.addNameSpace(prefixAtom, xmlNamespaces[idx].uri);
        }

        serializer.QueryInterface(Components.interfaces.nsIRDFXMLSource);
        serializer.Serialize(outputStream);
        this.mRDFXML = outputStream.data;
        this.mRDFXMLTainted = false;
        return this.mRDFXML;
    },

    close : function() {
        if (this.debug > 1) dump('* close()\n');
        this.mOpen = false;
        if (this.mRDFObserver) {
            this.mDS.RemoveObserver(this.mRDFObserver);
            this.mRDFObserver = null;
        }
        if (this.mXMLSink) {
            this.mDS.QueryInterface(Components.interfaces.nsIRDFXMLSink);
            this.mDS.removeXMLSinkObserver(this.mXMLSink);
            this.mDS.QueryInterface(Components.interfaces.nsIRDFDataSource);
            this.mXMLSink = null;
        }
        this.mDS = null;
        this.mStatements = null;
    },

    open : function(aDataSource) {
        if (this.debug > 1) dump('= RDFInspector.open()\n');
        if (!aDataSource instanceof Components.interfaces.nsIRDFDataSource) {
            var uri = aDataSource;
            if (this.debug > 2) dump('  + uri = "' + uri + '"\n');
            aDataSource = RDF.GetDataSource(uri);
        }
        if (this.debug > 2) dump('  + aDataSource = "' + aDataSource + '"\n');

        // Pre-load the datasource metadata
        this.mRDFXMLTainted = true;
        this.mDS = aDataSource;
        this.mURI = aDataSource.URI;
        if (aDataSource instanceof Components.interfaces.nsIRDFCompositeDataSource)
            this.mDSType = "nsIRDFCompositeDataSource";
        else if (aDataSource instanceof Components.interfaces.nsIRDFRemoteDataSource)
            this.mDSType = "nsIRDFRemoteDataSource";
        else if (aDataSource instanceof Components.interfaces.nsIRDFInMemoryDataSource)
            this.mDSType = "nsIRDFInMemoryDataSource";
        else if (aDataSource instanceof Components.interfaces.nsIRDFPropagatableDataSource)
            this.mDSType = "nsIRDFPropagatableDataSource";
        else if (aDataSource instanceof Components.interfaces.nsIRDFPurgeableDataSource)
            this.mDSType = "nsIRDFPurgeableDataSource";
        else if (aDataSource instanceof Components.interfaces.nsIRDFDataSource)
            this.mDSType = "nsIRDFDataSource";
        else
            this.mDSType = "Unknown";

        if (this.debug > 2) dump('  + this.mDSType = "' + this.mDSType + '"\n');

        // Initialize our statement storage array
        this.mStatements = new Array();
        if (this.debug > 2) dump('  + this.mStatements = "' + this.mStatements + '"\n');

        // Initialize our children storage array
        this.mChildren = new RDFInspectorCollection();
        if (this.debug > 2) dump('  + this.mChildren = "' + this.mChildren + '"\n');

        // Add the RDF statement observer
        var parent = this;
        this.mRDFObserver = {
            onAssert : function(aDataSource, aSource, aProperty, aTarget) {
                if (this.debug > 1) dump('+ RDFInspector.mRDFObserver.onAssert("' + aDataSource + '", "' + aSource.Value + '", "' + aProperty.Value + '", "' + aTarget + '")\n');
                parent.addStatement(aSource, aProperty, aTarget);
                if (this.debug > 3) dump('  * Added the statement\n');
            },

            onUnassert : function(aDataSource, aSource, aProperty, aTarget) {
                if (this.debug > 1) dump('+ RDFInspector.mRDFObserver.onUnassert("' + aDataSource + '", "' + aSource.Value + '", "' + aProperty.Value + '", "' + aTarget + '")\n');
            },

            onChange : function(aDataSource, aSource, aProperty, aOldTarget, aNewTarget) {
                if (this.debug > 1) dump('+ RDFInspector.mRDFObserver.onChange("' + aDataSource + '", "' + aSource.Value + '", "' + aProperty.Value + '", "' + aOldTarget + '", "' + aNewTarget + '")\n');
            },
            onMove : function(aDataSource, aOldSource, aNewSource, aProperty, aTarget) {
                if (this.debug > 1) dump('+ RDFInspector.mRDFObserver.onMove("' + aDataSource + '", "' + aOldSource.Value + '", "' + aNewSource.Value + '", "' + aProperty.Value + '", "' + aTarget + '")\n');
            },
            onBeginUpdateBatch : function(aDataSource) {
                if (this.debug > 1) dump('+ RDFInspector.mRDFObserver.onBeginUpdateBatch()\n');
            },
            onEndUpdateBatch : function(aDataSource) {
                if (this.debug > 1) dump('+ RDFInspector.mRDFObserver.onEndUpdateBatch()\n');
            }
        };
        aDataSource.AddObserver(this.mRDFObserver);

        // Create the XML Sink to watch the file load
        if (aDataSource instanceof Components.interfaces.nsIRDFRemoteDataSource) {
            if (this.debug > 3) dump('  * This is a remote datasource\n');
            if (!aDataSource.loaded && aDataSource instanceof Components.interfaces.nsIRDFXMLSink) {
                if (this.debug > 3) dump('  * Adding an XML sink to this datasource\n');
                if (this.debug > 4) dump('  * 1\n');
                this.mXMLSink = {
                    onBeginLoad : function(sink) {
                        if (parent.debug > 1) dump('+ RDFInspector.mXMLSink.onBeginLoad("' + sink + '")\n');
                        parent.mLoading = true;
                        if (parent.debug > 2) dump('  + parent.mLoading = "' + parent.mLoading + '"\n');
                    },
                    onInterrupt : function(sink) {
                        if (parent.debug > 1) dump('+ RDFInspector.mXMLSink.onInterrupt("' + sink + '")\n');
                    },
                    onResume : function(sink) {
                        if (parent.debug > 1) dump('+ RDFInspector.mXMLSink.onResume("' + sink + '")\n');
                    },
                    onError : function(sink,status,msg) {
                        if (parent.debug > 1) dump('+ RDFInspector.mXMLSink.onError("' + sink + '", "' + status + '", "' + msg + '")\n');
                    },
                    onEndLoad : function(sink) {
                        if (parent.debug > 1) dump('+ RDFInspector.mXMLSink.onEndLoad("' + sink + '")\n');
                        sink.removeXMLSinkObserver(this);
                        sink.QueryInterface(Components.interfaces.nsIRDFDataSource);

                        parent.mLoading = false;
                        if (parent.debug > 2) dump('  + parent.mLoading = "' + parent.mLoading + '"\n');

                        parent.mXMLSink = null;
                    }
                };
                if (this.debug > 4) dump('  * 2\n');
                aDataSource.QueryInterface(Components.interfaces.nsIRDFXMLSink);
                if (this.debug > 4) dump('  * 3\n');
                aDataSource.addXMLSinkObserver(this.mXMLSink);
                if (this.debug > 4) dump('  * 4\n');
            } else {
                if (this.debug > 3) dump('  * Fetching the statements manually\n');
                this.importStatements();
                if (this.debug > 4) dump('  * Done fetching statements\n');
            }
        }

        // Suck up the contents if we have a composite datasource
        else if (aDataSource instanceof Components.interfaces.nsIRDFCompositeDataSource) {
            var dsEnum = aDataSource.GetDataSources();
            if (this.debug > 2) dump('  + dsEnum = "' + dsEnum + '"\n');
            while (dsEnum.hasMoreElements()) {
                var ds = dsEnum.getNext();
                if (this.debug > 2) dump('  + ds = "' + ds + '"\n');
                if (ds instanceof Components.interfaces.nsIRDFDataSource) {
                    if (this.debug > 2) dump('  + ds.URI = "' + ds.URI + '"\n');
                    if (this.debug > 3) dump('  * Fetching the statements manually\n');
                    try {
                        var dsObj = new RDFInspector(ds);
                        if (this.debug > 2) dump('  + dsObj = "' + dsObj + '"\n');
                        this.mChildren.addInspector(dsObj);
                        try {
                            this.importStatements(ds);
                        } catch (e) {
                            if (this.debug > 0) dump('  Exception occurred while importing statements: ' + e + '\n');
                        }
                        if (this.debug > 4) dump('  * Done fetching statements\n');
                    } catch (e) {
                        dump('!!! An exception occurred while loading a child RDF Inspector: ' + e + '\n');
                    }
                }
            }
        }

        // Suck up the contents if we have something else
        else {
            if (this.debug > 3) dump('  * Fetching the statements manually\n');
            this.importStatements();
            if (this.debug > 4) dump('  * Done fetching statements\n');
        }
    },

    cloneDataSource : function() {
        if (this.debug > 1) dump('= RDFInspector.cloneDataSource()\n');

        var outputDS = Components
            .classes["@mozilla.org/rdf/datasource;1?name=in-memory-datasource"]
            .createInstance(Components.interfaces.nsIRDFDataSource);
        for (var idx = 0; idx < this.mStatements.length; idx++) {
            var statement = this.mStatements[idx];
            outputDS.Assert(statement.source, statement.property, statement.target, true);
        }
        return outputDS;
    },

    getStatementByIndex : function(aIndex) {
        //if (this.debug > 1) dump('+ RDFInspector.getStatementByIndex("' + aIndex + '")\n');
        return this.mStatements[aIndex];
    },

    addStatement : function(aSource, aProperty, aTarget) {
        if (this.debug > 1) dump('+ RDFInspector.addStatement("' + aSource.Value + '", "' + aProperty.Value + '", "' + aTarget + '")\n');
        var element = {
            source   : aSource,
            property : aProperty,
            target   : aTarget,
        };
        this.mStatements.push(element);
        if (this.mRDFXMLTainted == false) this.mRDFXMLTainted = true;
        if (this.debug > 3) dump('  * Statements list is now ' + this.mStatements.length + ' elements long\n');
    },

    importStatements : function(ds) {
        if (this.debug > 1) dump('= RDFInspector.importStatements("' + ds + '")\n');
        if (!ds) {
            ds = this.mDS.QueryInterface(Components.interfaces.nsIRDFDataSource);
        }
        if (this.debug > 2) dump('  + ds = "' + ds + '"\n');
        if (this.debug > 2) dump('  + ds.URI = "' + ds.URI + '"\n');

        var resourceEnum = ds.GetAllResources();
        if (this.debug > 2) dump('  + resourceEnum = "' + resourceEnum + '"\n');
        while (resourceEnum.hasMoreElements()) {
            var resource = resourceEnum.getNext().QueryInterface(Components.interfaces.nsIRDFResource);
            if (this.debug > 2) dump('  + resource.Value = "' + resource.Value + '"\n');

            var predicateEnum = ds.ArcLabelsOut(resource);
            if (this.debug > 2) dump('  + predicateEnum = "' + predicateEnum + '"\n');
            while (predicateEnum.hasMoreElements()) {
                var predicate = predicateEnum.getNext().QueryInterface(Components.interfaces.nsIRDFResource);
                if (this.debug > 2) dump('  + predicate.Value = "' + predicate.Value + '"\n');

                var objectEnum;
                try {
                    objectEnum = ds.GetTargets(resource, predicate, true);
                } catch (e) {
                    if (this.debug > 3) dump('  * Caught an exception while getting predicate targets: ' + e + '.\n');
                }
                if (this.debug > 2) dump('  + objectEnum = "' + objectEnum + '"\n');
                while (objectEnum.hasMoreElements()) {
                    var object;
                    try {
                        object = objectEnum.getNext();
                    } catch (e) {
                        if (this.debug > 3) dump('  * Caught an exception while processing the object enumerator: ' + e + '.\n');
                    }

                    // Add the statement count for this datasource
                    object = this.deriveNodeType(object);
                    if (this.debug > 2) dump('  + object = "' + object + '"\n');
                    this.addStatement(resource, predicate, object);
                }
            }
        }
        if (this.debug > 3) dump('  * Finished processing resources\n');
    },

    deriveNodeType : function(node) {
        if (node instanceof Components.interfaces.nsIRDFResource) {
            node = node.QueryInterface(Components.interfaces.nsIRDFResource);
        } else if (node instanceof Components.interfaces.nsIRDFLiteral) {
            node = node.QueryInterface(Components.interfaces.nsIRDFLiteral);
        } else if (node instanceof Components.interfaces.nsIRDFDate) {
            node = node.QueryInterface(Components.interfaces.nsIRDFDate);
        } else if (node instanceof Components.interfaces.nsIRDFInt) {
            node = node.QueryInterface(Components.interfaces.nsIRDFInt);
        } else if (node instanceof Components.interfaces.nsIRDFBlob) {
            node = node.QueryInterface(Components.interfaces.nsIRDFBlob);
        } else {
            node = node.QueryInterface(Components.interfaces.nsIRDFNode);
        }
        return node;
    },
};

function RDFInspectorCollection() {
    this.mInspectors = new Array();
    this.mObservers = new Array();
}
RDFInspectorCollection.prototype = {
    mInspectors : null,
    mObservers : null,
    get length() { return this.mInspectors.length },

    AddObserver : function(aObserver) {
        if (DEBUG > 1) dump('= RDFInspectorCollection.AddObserver("' + aObserver + '")\n');
        this.mObservers.push(aObserver);
    },

    addInspector : function(aInspector) {
        if (DEBUG > 1) dump('= RDFInspectorCollection.addInspector("' + aInspector + '")\n');
        this.mInspectors.push(aInspector);
        for (var idx in this.mObservers) {
            try {
                this.mObservers[idx].onAddInspector(aInspector);
            } catch (e) {}
        }
    },

    removeInspector : function(aInspector) {
        if (DEBUG > 1) dump('= RDFInspectorCollection.removeInspector("' + aInspector + '")\n');
        for (var i = 0; i < this.mInspectors.length; i++) {
            if (aInspector == this.mInspectors[i]) {
                if (DEBUG > 4) dump('  * Removing index ' + i + '\n');
                var result = this.mInspectors.slice(i, 1);
                for (var idx in this.mObservers) {
                    try {
                        this.mObservers[idx].onRemoveInspector(aInspector);
                    } catch (e) {}
                }
                return result[0];
            }
        }
        return null;
    },

    removeItem : function(aIndex) {
        if (DEBUG > 1) dump('= RDFInspectorCollection.removeItem("' + aIndex + '")\n');
        var result = this.mInspectors.slice(aIndex, 1);
        for (var idx in this.mObservers) {
            try {
                this.mObservers[idx].onRemoveInsector(result[0]);
            } catch (e) {}
        }
        return result[0];
    },

    item : function(aIndex) {
        //if (DEBUG > 1) dump('= RDFInspectorCollection.item("' + aIndex + '")\n');
        return this.mInspectors[aIndex];
    },
};

function RDFInspectorTreeView() {}
RDFInspectorTreeView.prototype = {
    // default values
    mRowCount: null,

    get wrappedJSObject() { return this; },

    // nsISupports
    QueryInterface : function(aIID) {
        if (Components.interfaces.nsIClassInfo.equals(aIID))
            return RDFInspectorTreeView.prototype;
        if (Components.interfaces.nsITreeView.equals(aIID) ||
                Components.interfaces.nsISupportsWeakReference.equals(aIID) ||
                Components.interfaces.nsISupports.equals(aIID))
            return this;
        throw 0x80004002; // Components.results.NS_NOINTERFACE;
    },

    // nsIClassInfo
    getInterfaces: function(count) {
        count.value = 4;
        return [
            Components.interfaces.nsITreeView,
            Components.interfaces.nsIClassInfo, 
            Components.interfaces.nsISupportsWeakReference, 
            Components.interfaces.nsISupports
        ];
    },
    getHelperForLanguage: function(language) { return null; },
    get contractID() { return null; },
    get classDescription() { return "nsTreeView"; },
    get classID() { return null; },
    get implementationLanguage() { return Components.interfaces.nsIProgrammimgLanguage.JAVASCRIPT; },
    get flags() { return Components.interfaces.nsIClassInfo.DOM_OBJECT; },

    // nsITreeView
    get rowCount() {
        if (DEBUG > 1) dump('= get RDFInspectorTreeView.rowCount()\n');
        return this.mRowCount;
    },
    set rowCount(aCount) {
        if (DEBUG > 1) dump('= set RDFInspectorTreeView.rowCount("' + aCount + '")\n');
        if (DEBUG > 3) dump('  * 1\n');
        this.mRowCount = aCount;
        if (DEBUG > 3) dump('  * 2\n');
        if (this.treeBox) {
            try {
                this.treeBox.rowCountChanged(1, aCount);
                if (DEBUG > 3) dump('  * Updated row count\n');
            } catch (e) {
                if (DEBUG > 0) dump('  Exception occurred while updating the tree box row count: ' + e + '\n');
            }
        }
    },

    // Basic functions
    selection: null,
    treeBox: null,
    setTree:             function(treeBox) { this.treeBox = treeBox; if (!treeBox) this.selection = null; },

    // Display Properties
    getRowProperties:    function(row, prop) { },
    getCellProperties:   function(row, column, prop) { },
    getColumnProperties: function(column, elem, prop) { },
    isSeparator:         function(row) { return false; },
    isEditable:          function(row, column) { return false; },

    // Heirarchial Functions
    isContainer:         function(row) { },
    isContainerOpen:     function(row) { },
    isContainerEmpty:    function(row) { return false; },
    toggleOpenState:     function(row) { },
    getParentIndex:      function(row) { },
    hasNextSibling:      function(row, after) { },
    getLevel:            function(row) { return 0; },

    // Row contents functions
    getImageSrc:         function(row, column) { },
    getProgressMode :    function(row,column) { },
    getCellValue:        function(row, column) { },
    getCellText:         function(row, column) { },
    cycleCell:           function(row, column) { },

    // Sorting
    isSorted:            function() { return false; },
    cycleHeader:         function(col, elem) { },

    // Selection
    selectionChanged:    function() { },

    // Drag & Drop
    canDropOn:           function(row) { return false; },
    canDropBeforeAfter:  function(row, before) { return false; },
    drop:                function(row, orientation) { },

    // Commands
    performAction:       function(action) { },
    performActionOnCell: function(action, row, column) { },
};

// DataSource Tree View, wrapping a collection
RDFInspectorDataSourceTreeView.prototype = new TreeView();
RDFInspectorDataSourceTreeView.prototype.constructor = RDFInspectorDataSourceTreeView;
RDFInspectorDataSourceTreeView.superclass = TreeView.prototype;
function RDFInspectorDataSourceTreeView(aCollection) {
    if (DEBUG > 1) dump('= RDFInspectorDataSourceTreeView("' + aCollection + '")\n');
    if (arguments.length > 0) {
        this.init();
        this.mCollection = aCollection;
        aCollection.AddObserver(this);
        if (DEBUG > 1) dump("Top is " + this.top + "\n");
        for (var idx = 0; idx < aCollection.length; idx++) {
            this.addInspector(aCollection.item(idx));
        }
    }
}
RDFInspectorDataSourceTreeView.prototype.addInspector = function(aInspector) {
    if (DEBUG > 1) dump('= RDFInspectorDataSourceTreeView.addInspector ("' + aInspector + '")\n');
    var item = this.appendItem(new RDFInspectorDataSourceTreeItem(aInspector));
    if (!aInspector.label)
        aInspector.label = strings.getString("defaultDatasourceLabel") + ' ' + this.top.rowCount;
    if (DEBUG > 1) dump("Top is: " + this.top + "\n");
    if (DEBUG > 1) dump("View length: " + this.top.rowCount + "\n");
};
RDFInspectorDataSourceTreeView.prototype.onAddInspector = function(aInspector) {
    if (DEBUG > 1) dump('= RDFInspectorDataSourceTreeView.onAddInspector ("' + aInspector + '")\n');
    this.addInspector(aInspector);
};

// DataSource Tree Item, wrapping an individual datasource
function RDFInspectorDataSourceTreeItem(aInspector) {
    if (DEBUG > 1) dump('= RDFInspectorDataSourceTreeItem("' + aInspector + '")\n');
    this.init();
    this.mInspector = aInspector;
    if (DEBUG > 1) dump("    this.mInspector.hasChildren = " + this.mInspector.hasChildren + "\n");
    if (this.mInspector.hasChildren) {
        if (DEBUG > 1) dump("This tree item has children; creating it\n");
        this.tree = new RDFInspectorDataSourceTreeView(this.mInspector.children);
    }
}
RDFInspectorDataSourceTreeItem.prototype = new TreeItem();
RDFInspectorDataSourceTreeItem.prototype.getCell = function(aCell) {
    if (DEBUG > 1) dump('= RDFInspectorDataSourceTreeItem.getCell("' + aCell + '")\n');
    // "Watch" column
    if (aCell == 'watchDSCol')
        return { imageSrc : 'chrome://rdfinspector/skin/icons/check.png' };

    // Label column
    else if (aCell == 'labelDSCol')
        return { text : this.mInspector.label };

    // URI column
    else if (aCell == 'uriDSCol')
        return { text : this.mInspector.URI };

    // Type column
    else if (aCell == 'typeDSCol')
        return { text : this.mInspector.type };

    // Size column
    else if (aCell == 'sizeDSCol')
        return { text : this.mInspector.length };
};
RDFInspectorDataSourceTreeItem.prototype.__defineGetter__("inspector", function() {
    if (DEBUG > 1) dump("get RDFInspectorDataSourceTreeItem.inspector :: this.mInspector = \"" + this.mInspector + "\"\n");
    return this.mInspector;
});

RDFInspectorDataSourceTreeView.prototype.onRemoveInspector = function(aInspector) {
    if (DEBUG > 1) dump('= RDFInspectorDataSourceTreeView.onRemoveInspector ("' + aInspector + '")\n');
    /* TODO: Make it properly remove this item
    this.rowCount--;
    if (DEBUG > 1) dump('  + this.rowCount = "' + this.rowCount + '"\n');
    this.treeBox.invalidate();
    if (DEBUG > 1) dump('  * Invalidated treeBox\n');
    */
};


function RDFInspectorStatementTreeView(aDS) {
    if (DEBUG > 1) dump('= RDFInspectorStatementTreeView("' + aDS + '")\n');
    this.mDS = aDS;
    this.rowCount = aDS.length;
}
RDFInspectorStatementTreeView.prototype = new RDFInspectorTreeView();
RDFInspectorStatementTreeView.prototype.getCellText = function(row, column) {
    //if (DEBUG > 1) dump('= RDFInspectorStatementTreeView.getCellText("' + row + '", "' + column + '")\n');
    var statement = this.mDS.getStatementByIndex(row);
    //if (DEBUG > 2) dump('  + statement = "' + statement + '"\n');

    var text = null;
    if (column == 'subjectStmtCol') {
        text = statement.source.Value;
        if (showNamespaces)
            text = convertURIToNS(text);
    }

    else if (column == 'predicateStmtCol') {
        text = statement.property.Value;
        if (showNamespaces)
            text = convertURIToNS(text);
    }

    else if (column == 'objectStmtCol') {
        text = statement.target.Value;
        if (statement.target instanceof Components.interfaces.nsIRDFResource && showNamespaces)
            text = convertURIToNS(text);
    }

    //if (DEBUG > 2) dump('  + text = "' + text + '"\n');
    return text;
};

function convertURIToNS(aURI) {
    for (var idx = 0; idx < xmlNamespaces.length; idx++) {
        if (!xmlNamespaces[idx].enabled) continue;
        if (aURI.indexOf(xmlNamespaces[idx].uri) == 0) {
            return xmlNamespaces[idx].prefix + ":" + aURI.substring(xmlNamespaces[idx].uri.length);
        }
    }
    return aURI;
}

window.updateNamespaceList = updateNamespaceList;
function updateNamespaceList() {
    xmlNamespaces = RDFInspectorSettings.GetAllNamespaces();
    gStatementTree.treeBoxObject.invalidate();
}

var gRDFCollection = new RDFInspectorCollection();
var gDatasourceTree, gStatementTree;
var dsTreeView;

function onLoad() {
    if (DEBUG > 1) dump('= onLoad()\n');

    // Initialize globals
    strings = document.getElementById("rdfinspectorStrings");

    // Initialize application state from preferences
    showNamespaces = RDFInspectorSettings.BoolPref("rdfinspector.showNamespaces");
    document.getElementById('bcast_tbViewOptions_mnUseNSPrefixes').setAttribute('checked', showNamespaces.toString());

    // Set the datasources up
    gDatasourceTree = document.getElementById("datasourceTree");
    gStatementTree = document.getElementById("statementList");
    updateNamespaceList();

    // Set up our datasource treeview
    dsTreeView = new RDFInspectorDataSourceTreeView(gRDFCollection);
    gDatasourceTree.view = dsTreeView;

    // Load any arguments to see if we need to open up a datasource right away
    if (window.arguments && window.arguments.length) {
        for (var i = 0; i < window.arguments.length; i++) {
            loadDataSource(window.arguments[i]);
        }
    }
}

window.loadDataSource = loadDataSource;
function loadDataSource(uri) {
    if (DEBUG > 1) dump('= loadDataSource("' + uri + '")\n');
    var ds;

    // Load the datasource
    if (uri instanceof Components.interfaces.nsIRDFDataSource) {
        ds = uri;
    } else {
        ds = RDF.GetDataSource(uri);
    }
    if (DEBUG > 2) dump('  + ds = "' + ds + '"\n');

    var dsObj = new RDFInspector(ds);
    if (DEBUG > 2) dump('  + dsObj = "' + dsObj + '"\n');

    gRDFCollection.addInspector(dsObj);
    if (DEBUG > 2) dump('  + gRDFCollection.length = "' + gRDFCollection.length + '"\n');
}

// EVENT HANDLERS
function openURI() {
    if (DEBUG > 1) dump('= openURI()\n');
    openDialog("chrome://rdfinspector/content/openURI.xul", "_blank", "chrome,modal,titlebar", window);
}

function newDataSource() {
    alert("This doesn't work yet; sorry.");
}

function saveAs() {
    const FILEPICKER_CTRID = "@mozilla.org/filepicker;1";
    const FILEOUT_CTRID = "@mozilla.org/network/file-output-stream;1";
    const nsIFilePicker = Components.interfaces.nsIFilePicker;
    const nsIFileOutputStream = Components.interfaces.nsIFileOutputStream;

    //alert("This doesn't work yet; sorry.");
    var dsItem = GetSelectedDatasource();
    if (dsItem == null) {
        // Whoa!  This shouldn't ever happen because we don't let them save if a datasource isn't selected
        // Oh well...chalk it up to bad validation, and let it pass silently.
        return;
    }
    var fp = Components
        .classes[FILEPICKER_CTRID]
        .createInstance(nsIFilePicker);

    fp.init(window, strings.getString("saveFileTitle"), nsIFilePicker.modeSave);
    fp.appendFilter(strings.getString("rdfFiles"), "*.rdf");
    fp.appendFilters(nsIFilePicker.filterXML);
    fp.appendFilters(nsIFilePicker.filterAll);
    var result = fp.show();
    if (result == nsIFilePicker.returnCancel)
        return;

    var ds = dsItem.cloneDataSource();
    var rdfResult = SerializeRDF(ds);

    var outputStream = Components
        .classes[FILEOUT_CTRID]
        .createInstance(nsIFileOutputStream);
    try {
        outputStream.init(fp.file, MODE_WRONLY | MODE_CREATE | MODE_TRUNCATE, 0664, 0);
        outputStream.write(rdfResult, rdfResult.length);
        outputStream.close();
    } catch (e) {
        promptService.alert(window, strings.getString("saveError.title"), strings.getString("saveError.message"));
        return;
    }
}

function closeDS() {
    if (DEBUG > 1) dump('= closeDS()\n');
    alert("This doesn't work yet; sorry.");
    return;
    var idx = gDatasourceTree.view.selection.currentIndex;
    if (DEBUG > 2) dump('  + idx = "' + idx + '"\n');
    gRDFCollection.removeItem(idx);
    if (DEBUG > 3) dump('  * Done closing\n');
}

function openFile() {
    if (DEBUG > 1) dump('= openFile()\n');
    var fp = Components
        .classes["@mozilla.org/filepicker;1"]
        .createInstance(Components.interfaces.nsIFilePicker);

    fp.init(window, strings.getString("openFileTitle"), Components.interfaces.nsIFilePicker.modeOpen);
    fp.appendFilter(strings.getString("rdfFiles"), "*.rdf");
    fp.appendFilters(Components.interfaces.nsIFilePicker.filterXML);
    fp.appendFilters(Components.interfaces.nsIFilePicker.filterAll);

    var result = fp.show();
    if (result == Components.interfaces.nsIFilePicker.returnOK) {
        var ioServ = Components
            .classes["@mozilla.org/network/io-service;1"]
            .getService(Components.interfaces.nsIIOService);

        var fph = ioServ
            .getProtocolHandler("file")
            .QueryInterface(Components.interfaces.nsIFileProtocolHandler);

        var fileURI = fph.getURLSpecFromFile(fp.file);

        if (DEBUG > 3) dump(" * Before loadDataSource\n");
        loadDataSource(fileURI);
        if (DEBUG > 3) dump(" * After loadDataSource\n");
    }
}

function updateRDFXML(event) {
    if (DEBUG > 1) dump('updateRDFXML("' + event + '")\n');
    /* TODO Optimize this.  We can check to see if the datasource has
     *      changed at all, and if not, don't re-build the XML string.
    */
    var textBox = document.getElementById("statementRDFXML");
    var dsItem = GetSelectedDatasource();

    if (dsItem == null) {
        textBox.value = "";
        return;
    }
    if (textBox.selectedDataSource == dsItem)
        return;

    textBox.selectedDataSource = dsItem;
    textBox.value = dsItem.serialize();
}

function openDatasourceRenameDialog(event) {
    if (DEBUG > 1) dump('openDatasourceRenameDialog("' + event + '")\n');
    // Fetch the datasource currently selected
    var dsItem = GetSelectedDatasource();
    if (dsItem == null) return;

    // Prompt them for a new label
    var result = prompt(strings.getString("renameLabel.prompt"), dsItem.label, strings.getString("renameLabel.title"), 0);
    if (result ==  null) return;

    // Set the new label, and let the treeview do it's thing
    dsItem.label = result;
}

function getSelectedURI(event) {
    if (DEBUG > 1) dump('= getSelectedURI("' + event + '")\n');
    var currentIndex = event.target.view.selection.currentIndex;
    if (currentIndex < 0) return null;
    return event.target.view.getResourceAtIndex(currentIndex).Value;
}

function selectDatasourceTree(event) {
    if (DEBUG > 1) dump('= selectDatasourceTree("' + event + '")\n');
    var dsItem = GetSelectedDatasource();
    if (dsItem == oldSelectedDataSource) {
        return;
    } else {
        oldSelectedDataSource = dsItem;
    }
    var bcastClose = document.getElementById("bcastCloseDS");

    if (dsItem == null) {
        bcastClose.setAttribute("disabled", "true");
        return;
    } else {
        bcastClose.setAttribute("disabled", "false");
    }

    var stmtTreeView = new RDFInspectorStatementTreeView(dsItem);
    gStatementTree.view = stmtTreeView;

    if (document.getElementById("statementPanel").selectedPanel.id == "tabpanelStatementRDFXML") {
        updateRDFXML(event);
    }
}

function GetSelectedDatasource() {
    if (DEBUG > 1) dump("GetSelectedDatasource()\n");
    var tree = document.getElementById('datasourceTree');
    if (DEBUG > 1) dump("GetSelectedDatasource() :: tree = \"" + tree + "\"\n");
    var idx = tree.view.selection.currentIndex;
    if (DEBUG > 1) dump("GetSelectedDatasource() :: idx = \"" + idx + "\"\n");
    if (idx == -1)
        return null;
    return dsTreeView.getItemByRow(idx).inspector;
}

function toggleNamespacePrefixes() {
    if (DEBUG > 1) dump('= toggleNamespacePrefixes()\n');
    showNamespaces = !showNamespaces;
    RDFInspectorSettings.BoolPref("rdfinspector.showNamespaces", showNamespaces);
    var bcaster = document.getElementById('bcast_tbViewOptions_mnUseNSPrefixes');
    bcaster.setAttribute('checked', showNamespaces.toString());
}

function openEditNamespacesDialog() {
    window.openDialog(
        'chrome://rdfinspector/content/editNamespace.xul',
        'rdfinspector-editNamespace',
        'chrome,width=450,height=400'
    );
}
